using System;
using System.Drawing;
using System.Drawing.Drawing2D;


namespace ShapedFormDataReader
{
	/// <summary>
	/// Parse through the data (.dat) file saved by VBSFC version 6 to produce a region.
	/// </summary>
	/// <remarks>
	/// This class was derived from the VB.NET data file reader created by John Risby
	/// (john@wildathome.com).  The main difference (other than the language) is that
	/// the C# datafile reader does not require the use of the VB legacy runtimes.  
	/// There should be a performance increase and the distribution will be smaller
	/// by avoiding the legacy runtimes.  The below summary is verbatim from the VB.NET 
	/// data file reader.
	/// 
	/// - MATT THEULE, 22Jan2003
	/// 
	///	This module is an example of how to read the .dat Data files
	///	saved by VBSFC version 6 to produce a region.
	///
	///The .dat binary file format is as follows:
	///The leftmost 5 bytes must be "VBSFC"
	///
	///The next 2 bytes are the Major and Minor versions of VBSFC that wrote the file.
	///The 8th and 9th bytes form an integer value specifying the largest number of points
	///in any polygon.  This is handy if you want to define an array to hold polygon points
	///and don't want to have to resize it.
	///
	///Subsequent bytes (10 and onwards) follow the fixed format:
	///ObjectType,Top,Left,Right,Bottom,RoundedRectX,RoundedRectY,CombinationMethod
	///Where ObjectType and CombinationMethod are single bytes and all the other are double
	///bytes (integers).  ObjectType and CombinationMethod values are defined in the Enum
	///structures below for ObjectTypes and Methods.  If any property is not used it
	///is 0, and still takes up the same 2 bytes, so each record is of the same length.
	///
	///In the case of a polygon object type the fixed length object bytes are followed by
	///an integer which specifies the length (in integers, so double this number of bytes)
	///of a variable length section containing integer points in the polygon, X followed by Y.
	///For polygons the Top property specifies Winding or Alternate fill.
	///
	///To apply the resulting region to a form, put the line:
	///   Region = ReadFormRegion(strFilename)
	///in the Form Load event.
	/// </remarks>
	/// 

	public class ShapedFormReader
	{
		private enum ObjectTypes
		{
			Rectangle = 0,
			Polygon = 1,
			RoundedRectangle = 2,
			Ellipse = 3
		}
		private enum Methods
		{
			ANDCombine = 0,
			ORCombine = 1,
			XORCombine = 2,
			NOTCombine = 3
		}
/// <summary>
/// 
/// </summary>
		public ShapedFormReader()
		{
			//
			// TODO: Add constructor logic here
			//

		}
		private System.IO.BinaryReader mShapeData;
		private string mFileErrorMessage = "";
		private void MoveToStartOfFixedLengthInfo()
		{
			//The info we are skipping is
			// version info and largest number of 
			// points in a polygon.
			mShapeData.ReadBytes(4);
		}
		private void FileGet (ref byte Value)
		{
			Value = mShapeData.ReadByte();
		}

		private void FileGet (ref short Value)
		{
			Value = mShapeData.ReadInt16();
		}
		private void FileGet (ref string Value)
		{
			System.Text.StringBuilder sB = new System.Text.StringBuilder();
			sB.Append(mShapeData.ReadChars(Value.Length));
			Value = sB.ToString();			
		}
		private bool FileOpen(string PathAndName)
		{	
			bool bReturn = false;
			try
			{
				mShapeData = new System.IO.BinaryReader(new System.IO.FileStream(PathAndName, System.IO.FileMode.Open, System.IO.FileAccess.Read));
				bReturn = true;
			}

			catch
			{
				//set the return value to false
				bReturn = false;
			}

			return bReturn;
		}
		private void FileClose()
		{
			mShapeData.Close();
		}
		private long Loc()
		{
			return mShapeData.BaseStream.Position;
		}
		private long LOF()
		{
			return mShapeData.BaseStream.Length;
		}

		private bool ValidFile (string strFilename, bool ignoreErrors)
		{
			try
			{
				string strBuffer;

				//Open the file
				if(FileOpen(strFilename))
				{

					strBuffer = "*****";    //Read 5 bytes
					FileGet(ref strBuffer);

					//Check for validity
					if (strBuffer == "VBSFC")
					{
						return true;
					}
					else
						//if the exception class (SFDFRException.cs) is not included
						//  comment out the check for ignoreErrors and just return false.
						//  No region will be generated, and the app will be in a default
						//  rectangle.
					{
						if(ignoreErrors)
						{
							mFileErrorMessage = "The selected file is not in the expected format.";
							return false;
						}
						else
						{	
							throw new ShapedFormDataReader.ShapedFormReaderFileFormatException(strFilename);					
						}
					}
				}
				else
				{
					//if the exception class (SFDFRException.cs) is not included
					//  comment out the check for ignoreErrors and just return false.
					//  No region will be generated, and the app will be in a default
					//  rectangle.
					if(ignoreErrors)
					{
						mFileErrorMessage = "The selected file cannot be found.";
						return false;
					}
					else
					{	
						throw new System.IO.FileNotFoundException("The Shaped Form data file could not be found.",strFilename);
					}
				}
			}
			catch (Exception exp)
			{
				throw exp;
			}
		}
		/// <summary>
		/// Reads the region information from the specified data file.
		/// </summary>
		/// <param name="strFilename">Full path and file name of the data file.</param>
		/// <param name="IgnoreFileErrors">Throw or suppress exceptions generated by finding and opening the datafile.
		/// <p>If errors are ignored, the app will continue without applying region information.  The text of the error can be found in the FileError property.</p></param>
		/// <returns>A region that can be applied to the form or to a specific control.</returns>
		public System.Drawing.Region ReadFormRegion(string strFilename, bool IgnoreFileErrors)
		{	
			try
			{
				return GetFormRegion(strFilename, IgnoreFileErrors);
			}
			catch (Exception e)
			{
				throw e;
			}
			
		}
		/// <summary>
		/// Reads the region information from the specified data file.
		/// </summary>
		/// <param name="strFilename">Full path and file name of the data file.</param>
		/// <returns>A region that can be applied to the form or to a specific control.</returns>
		public System.Drawing.Region ReadFormRegion(string strFilename)
		{
			try
			{
				return GetFormRegion(strFilename, false);
			}
			catch (Exception e)
			{
				throw e;
			}			
		}
		private System.Drawing.Region GetFormRegion(string strFilename, bool IgnoreFileErrors)
		{
			short PolyCount = 0;
			byte byteBuffer = 0;
			short intBuffer = 0;
			ObjectTypes ObjectType;
			short Top; 
			short Left;
			short Right;
			short Bottom;
			short RoundedRectX;
			short RoundedRectY;
			Methods Method;
			short Counter;
			Boolean bFirst;
			System.Drawing.Region rgnResult;
			GraphicsPath gPath;
			System.Drawing.PointF[] PolyPoints;

			try
			{
				rgnResult = new System.Drawing.Region();
				gPath = new GraphicsPath();

				bFirst = true;

				if (ValidFile(strFilename, IgnoreFileErrors))
				{
					//Create Polygon Points buffer
					MoveToStartOfFixedLengthInfo();

					do
					{
						FileGet(ref byteBuffer) ; ObjectType = (ObjectTypes)Enum.Parse(typeof(ObjectTypes),Enum.GetName(typeof(ObjectTypes),(int)byteBuffer),true);
						FileGet(ref intBuffer) ; Top = intBuffer;
						FileGet(ref intBuffer) ; Left = intBuffer;
						FileGet(ref intBuffer) ; Right = intBuffer;
						FileGet(ref intBuffer) ; Bottom = intBuffer;
						FileGet(ref intBuffer) ; RoundedRectX = intBuffer;
						FileGet(ref intBuffer) ; RoundedRectY = intBuffer;
						FileGet(ref byteBuffer) ; Method = (Methods)Enum.Parse(typeof(Methods),Enum.GetName(typeof(Methods),(int)byteBuffer),true);
				
						switch (ObjectType)
						{
							case ObjectTypes.Rectangle:
								gPath = new GraphicsPath();
								gPath.AddRectangle(new Rectangle(Left, Top, Right - Left, Bottom - Top));
								if (bFirst)
								{
									bFirst = false;
									rgnResult = new System.Drawing.Region(gPath);
								}
								else
								{
									AddPathToRegion(ref rgnResult, gPath, Method);
								}
								gPath.Dispose();
								break;
							case ObjectTypes.Polygon:
								gPath = new GraphicsPath();
								FileGet(ref PolyCount);
								PolyPoints = new System.Drawing.PointF[PolyCount];
								
								if (PolyCount > 0)
								{
									for (Counter = 0; Counter < PolyCount; Counter++)
									{
										FileGet(ref intBuffer) ; PolyPoints[Counter].X = intBuffer;
										FileGet(ref intBuffer) ; PolyPoints[Counter].Y = intBuffer;
									}
		
									switch (Top)
									{
										case 1:
											gPath.FillMode = FillMode.Alternate;
											break;
										case 2:
											gPath.FillMode = FillMode.Winding;
											break;
									}
									gPath.AddPolygon(PolyPoints);
									gPath.CloseFigure();
									if (bFirst)
									{
										bFirst = false;
										rgnResult = new System.Drawing.Region(gPath);
									}
									else
									{
										AddPathToRegion(ref rgnResult, gPath, Method);
									}
									
								}
								gPath.Dispose();
								break;
							case ObjectTypes.RoundedRectangle:
								gPath = new GraphicsPath();
								gPath = RoundedRect(Left, Top, (Right - Left), Bottom - Top, RoundedRectX, RoundedRectY);

								if (bFirst)
								{
									bFirst = false;
									rgnResult = new System.Drawing.Region(gPath);
								}
								else
								{
									AddPathToRegion(ref rgnResult, gPath, Method);
								}

								gPath.Dispose();
								break;
							case ObjectTypes.Ellipse:
								gPath = new GraphicsPath();
								gPath.AddEllipse(new Rectangle(Left, Top, (Right - Left) - 1, (Bottom - Top) - 2));
								if (bFirst)
								{
									bFirst = false;
									rgnResult = new Region(gPath);
								}
								else
								{
									AddPathToRegion(ref rgnResult, gPath, Method);
								}
								gPath.Dispose();
								break;
						}
					}while (Loc() != LOF());
					FileClose();
				
					return rgnResult;
				}
				else
				{
					return null;
				}
			}
			catch (Exception exp)
			{
				throw exp;
			}
		}
		private void AddPathToRegion(ref System.Drawing.Region rgnResult, GraphicsPath gPath, Methods Method)
		{
			try
			{			
				switch (Method)
				{
					case Methods.ANDCombine:
						rgnResult.Intersect(gPath);
						break;
					case Methods.NOTCombine:
						rgnResult.Exclude(gPath);
						break;
					case Methods.ORCombine:
						rgnResult.Union(gPath);
						break;
					case Methods.XORCombine:
						rgnResult.Xor(gPath);
						break;
				}
			}
			catch (Exception exp)
			{
				throw exp;
			}
		}
		private GraphicsPath RoundedRect(
			int x, int y,
			int width, int height, 
			int xrad, int yrad)
		{
			try
			{			
				GraphicsPath gPath;
				int cornerwidth;
				int cornerheight;

				width -= 1;//width - 1;
				height += 1; //height + 1;

				cornerwidth = xrad * 2;
				cornerheight = yrad * 2;

				gPath = new GraphicsPath();
				gPath.AddArc(x, y, cornerwidth, cornerheight, 180, 90);
				gPath.AddArc(x + width - cornerwidth, y, cornerwidth, cornerheight, 270, 90);
				gPath.AddArc(x + width - cornerwidth, y + height - cornerheight, cornerwidth, cornerheight, 0, 90);
				gPath.AddArc(x, y + height - cornerheight, cornerwidth, cornerheight, 90, 90);
				gPath.CloseFigure();
				return gPath;
			}
			catch (Exception exp)
			{
				throw exp;
			}
		}
		/// <summary>
		/// If exceptions are ignored, some information is available.
		/// </summary>
		public string FileError
		{
			get 
			{
				return mFileErrorMessage;
			}
		}

	}
}